import logging

import click

from fxq.ae.agent.constants import LOGGING_FORMAT
from fxq.ae.agent.fxq_ae_agent_app import app


@click.command()
@click.option('--debug', is_flag=True, help="Enable debug Logging for the application")
def main(debug: bool):
    logging.basicConfig(format=LOGGING_FORMAT, level=(logging.DEBUG if debug else logging.INFO))
    app.run(host="0.0.0.0")


if __name__ == '__main__':
    main()



import docker
import requests
from flask import Flask, request, jsonify
from fxq.core.beans.factory.annotation import Autowired

from fxq.ae.agent.model.job import Job
from fxq.ae.agent.model.run import Run
from fxq.ae.agent.model.status import Health
from fxq.ae.agent.service.consul import ConsulService
from fxq.ae.agent.service.docker import DockerService
from fxq.ae.agent.service.job import JobService

app = Flask(__name__)

consul_service = Autowired(type=ConsulService)

job_service = Autowired(type=JobService)

docker_service = Autowired(type=DockerService)


@app.route('/api/request', methods=['POST'])
def start():
    job: Job = Job.of_dict(request.json)
    run: Run = job_service.process_request(job)
    return jsonify(
        run.to_dict()
    )


@app.route('/api/actuator/health', methods=['GET'])
def check_health():
    try:
        docker_service.list_containers()
        return jsonify(Health("UP").to_dict())
    except requests.exceptions.ConnectionError as e:
        return jsonify(Health("DOWN", {"error": "Docker ConnectionError, Check Docker Engine and Socket is up"}).to_dict())
    except docker.errors.APIError:
        return jsonify(Health("DOWN", {"error": "Docker APIError, Check Docker Engine is Running and API Socket is working"}).to_dict())
    except Exception as e:
        print(e.__class__)
        return jsonify(Health("DOWN", {"error": str(e)}).to_dict())



import logging
from datetime import datetime

from docker.models.containers import Container
from fxq.core.beans.factory.annotation import Autowired
from fxq.core.stereotype import Service

from fxq.ae.agent.model.run import Run, Step, Command
from fxq.ae.agent.model.status import Status
from fxq.ae.agent.service.docker import DockerService

LOGGER = logging.getLogger(__name__)


@Service
class RunService:

    @Autowired
    def __init__(self, docker_service):
        self._docker_service: DockerService = docker_service

    def start(self, run: Run, workspace_path: str):
        LOGGER.info("Executing Run uuid %s" % run.uuid)
        try:
            self._do_run(run, workspace_path)
        except RunException:
            LOGGER.error("Run Failed, failing remaining Steps")
            self._abort_remaining_steps(run)
        LOGGER.info("Completed Executing the run %s" % run.uuid)

    def _do_run(self, run: Run, workspace_path):
        run.status = Status.IN_PROGRESS
        for step in run.steps:
            try:
                self._do_step(step, workspace_path)
            except Exception as e:
                LOGGER.error("Error occurred in Step #%s - %s, failing run, set to Debug for the full stacktrace" % (
                    step.number, step.name))
                LOGGER.debug(e)
                run.ended = datetime.now()
                run.status = Status.FAILED
                break

        if run.status == Status.IN_PROGRESS:
            run.ended = datetime.now()
            run.status = Status.SUCCESSFUL
        elif run.status == Status.FAILED:
            raise RunException("Run aborted due to Failed Run Status")

    def _do_step(self, step: Step, workspace_path):
        LOGGER.info("Setting up step %s" % step.name)
        step.status = Status.IN_PROGRESS
        container = None
        try:
            container = self._do_step_setup(step, workspace_path)
            self._do_step_work(step, container)
        except Exception as e:
            LOGGER.error(
                "Exception caught during step, marking as failed, set to Debug for the full stacktrace")
            LOGGER.debug(e)
            step.status = Status.FAILED
        finally:
            self._docker_service.teardown(container)

        if step.status == Status.IN_PROGRESS:
            step.status = Status.SUCCESSFUL
        elif step.status == Status.FAILED:
            raise StepException("Step aborted due to Failed Step Status")

    def _do_step_setup(self, step: Step, workspace_path: str):
        container = self._docker_service.provision(
            step.run.uuid,
            step.image,
            workspace_path
        )
        return container

    def _do_step_work(self, step: Step, container: Container):
        LOGGER.info("Running commands for step %s" % step.name)
        for command in step.commands:
            self._do_command(command, container)

    def _do_command(self, command: Command, container):
        LOGGER.info("Running command %s" % command.instruction)
        output = self._docker_service.run_command(container, command.instruction)
        while True:
            try:
                std_out, std_err = next(output)
                if std_out:
                    command.add_output(std_out.decode())
                if std_err:
                    command.add_error(std_err.decode())
                    raise StdErrException("Execution stopped due to command receiving stdErr")
            except StopIteration:
                break

    def _abort_remaining_steps(self, run: Run):
        for step in run.steps:
            if step.status == Status.PENDING:
                step.status = Status.ABORTED


class RunException(Exception):
    pass


class StepException(Exception):
    pass


class StdErrException(Exception):
    pass



from enum import Enum
from typing import Dict


class Status(Enum):
    PENDING = 1
    SUCCESSFUL = 2
    FAILED = 3
    IN_PROGRESS = 4
    PAUSED = 5
    STOPPED = 6
    ABORTED = 7


class Health():
    def __init__(self, status: str, details: Dict = None):
        self.status = status
        self.details = details

    def to_dict(self):
        resp = {
            'status': self.status
        }

        if self.details:
            resp["details"] = self.details

        return resp



import json
import logging
import threading
from http import HTTPStatus

import requests
from fxq.core.beans.factory.annotation import Autowired

from fxq.ae.agent.constants import JSON_HEADERS, URI_LIST_HEADERS
from fxq.ae.agent.service.consul import ConsulService

LOGGER = logging.getLogger(__name__)

consul_service: ConsulService = Autowired(type=ConsulService)
callback_host = None


def get_callback_host():
    global callback_host
    try:
        callback_host = f'{consul_service.get_callback_host()}'
        LOGGER.info(f'Configuring callback host as {callback_host}')
    except Exception as e:
        LOGGER.error(e)
    finally:
        t = threading.Timer(30.0, get_callback_host)
        t.start()


get_callback_host()


def get_callback_url(ref_object):
    if ref_object.__class__.__name__ == "Run":
        return f'{callback_host}/api/data/runs'
    if ref_object.__class__.__name__ == "Step":
        return f'{callback_host}/api/data/steps'
    if ref_object.__class__.__name__ == "Command":
        return f'{callback_host}/api/data/commands'


def register_new_object(ref_object, callback_url):
    '''
    Registers the object with the callback URL by posting the to_dict()
    content of the referenced object to the callback_url.

    Once the response is received from the server links are added showing the object
    has previously been created.

    :param ref_object: The Object to be serialized for the endpoint
    :param callback_url: The Endpoint to receive the POST request
    :return: NA
    '''
    LOGGER.debug(f"Registering New Object: {ref_object} at {callback_url}")
    r = requests.post(
        "%s" % callback_url,
        data=json.dumps(ref_object.to_dict()),
        headers=JSON_HEADERS)
    if r.status_code == HTTPStatus.CREATED:
        ref_object._links = r.json()["_links"]
    else:
        LOGGER.error("Post Callback Failed with Status Code %s" % r.status_code)
        LOGGER.debug("%s - %s" % (r.status_code, r.text))
        raise Exception("Post Callback Failed with Status Code %s" % r.status_code)

    update_links(ref_object)


def update_existing_object(ref_object, callback_url):
    '''
    Simply Sends a PATCH request with the serialized representation of the object to update assocaited
    fields on their change.

    :param ref_object: The Object to be serialized for the endpoint
    :param callback_url: The Endpoint to receive the POST request
    :return: NA
    '''
    LOGGER.debug(f"Updating Existing Object: {ref_object} at {ref_object._links['self']['href']}")
    r = requests.patch(
        ref_object._links["self"]["href"],
        data=json.dumps(ref_object.to_dict()),
        headers=JSON_HEADERS)
    if r.status_code != HTTPStatus.OK:
        LOGGER.error("Post Callback Failed with Status Code %s" % r.status_code)
        LOGGER.debug("%s - %s" % (r.status_code, r.text))
        raise Exception("Post Callback Failed with Status Code %s" % r.status_code)


def update_links(ref_object):
    '''
    Scans the _links definition as returned by the server, if it finds an attribute
    of the same name it will get the href from that linked attribute and associate it
    with this object.

    i.e. If ObjectA has links to self, objectA and objectB, this will look for an attribute
    by the name of objectB in the python objectA, it then get its link and uses a PUT request
    to associate the two together server side.

    :param ref_object: The object to scan for links and references
    :return: NA
    '''
    _link_exclusion = ['self', str(ref_object.__class__.__name__).lower()]
    for link_name, link_value in ref_object._links.items():
        if link_name not in _link_exclusion:
            try:
                associated_object = getattr(ref_object, link_name)
                LOGGER.info(
                    "%s is being linked to %s" % (link_value["href"], associated_object._links["self"]["href"]))
                requests.put(link_value["href"], data=associated_object._links["self"]["href"],
                             headers=URI_LIST_HEADERS)
            except AttributeError:
                LOGGER.info("No associated object exists of type %s for type %s" % (
                    link_name, ref_object.__class__.__name__))


def do_callback(ref_object):
    callback_url = get_callback_url(ref_object)
    if callback_url:
        LOGGER.info(f"Performing Callback for {ref_object.__class__.__name__} - {ref_object.to_dict()}")
        if ref_object._links:
            update_existing_object(ref_object, callback_url)
        elif ref_object._links is None:
            register_new_object(ref_object, callback_url)
    else:
        LOGGER.warning("No Callback URL could be determined")



import tempfile

PIPELINE_BASE = "%s/fxquants-pipelines" % tempfile.gettempdir()
PIPELINE_MOUNT_TARGET = "/opt/fxquants/pipeline"
PROJECTS_FOLDER = "projects"
PIPELINE_YML_NAME = "fxq-pipeline.yml"
LOGGING_FORMAT = "[%(asctime)s] [%(levelname)s] [%(name)s] - %(message)s"
JSON_HEADERS = {'content-type': 'application/json'}
URI_LIST_HEADERS = {'content-type': 'text/uri-list'}
ANALYTICS_SERVICE_ID = 'analytics-8200'



import uuid
from datetime import datetime
from typing import List, Dict

from fxq.ae.agent.callback.handler import do_callback
from fxq.ae.agent.model.job import Job
from fxq.ae.agent.model.status import Status


class Run:
    def __init__(self, job):
        self.uuid: str = str(uuid.uuid4())
        self.job: Job = job
        self._status: Status = Status.PENDING
        self.started: datetime = datetime.now()
        self.ended: datetime = None
        self.steps: List[Step] = []
        self._links: Dict = None
        do_callback(self)

    def to_dict(self):
        dict = {
            'uuid': self.uuid,
            'status': self.status.name,
            'started': self.started.isoformat(),
            'ended': self.ended.isoformat() if self.ended is not None else None
        }

        if self._links:
            dict["_links"] = self._links

        return dict

    @property
    def status(self):
        return self._status

    @status.setter
    def status(self, status):
        self._status = status
        do_callback(self)


class Step:
    def __init__(self, run, number, name, image):
        self.run: Run = run
        self.number: int = number
        self.name: str = name
        self.image: str = image
        self._status: Status = Status.PENDING
        self.commands: List[Command] = []
        self._links = None
        do_callback(self)

    def to_dict(self):
        return {
            'number': self.number,
            'name': self.name,
            'image': self.image,
            'status': self.status.name
        }

    @property
    def status(self):
        return self._status

    @status.setter
    def status(self, status):
        self._status = status
        do_callback(self)


class Command:
    def __init__(self, step, number, instruction):
        self.step: Step = step
        self.number: int = number
        self.instruction: str = instruction
        self.std_out: List[str] = []
        self.std_err: List[str] = []
        self._links = None
        do_callback(self)

    def add_output(self, output):
        self.std_out.append(output)
        do_callback(self)

    def add_error(self, error):
        self.std_err.append(error)
        do_callback(self)

    def to_dict(self):
        return {
            'number': self.number,
            'instruction': self.instruction,
            'stdOut': "".join(self.std_out),
            'stdErr': "".join(self.std_err)
        }



import logging

from fxq.ae.agent.constants import LOGGING_FORMAT
from fxq.ae.agent.fxq_ae_agent_app import app

logging.basicConfig(format=LOGGING_FORMAT, level=logging.getLogger('gunicorn.error').level)
application = app



import logging
import os
import shutil
import threading
import uuid

from fxq.core.beans.factory.annotation import Autowired
from fxq.core.stereotype import Service
from git import Repo

from fxq.ae.agent import constants
from fxq.ae.agent.factory.run import RunFactory
from fxq.ae.agent.model.job import Job
from fxq.ae.agent.model.run import Run
from fxq.ae.agent.service.docker import DockerService
from fxq.ae.agent.service.run import RunService

LOGGER = logging.getLogger(__name__)


@Service
class JobService:

    @Autowired
    def __init__(self, run_service, docker_service):
        self.run_service: RunService = run_service
        self.docker_service: DockerService = docker_service
        LOGGER.info("System Pipeline Base set to %s" % constants.PIPELINE_BASE)

    def process_request(self, job: Job) -> Run:
        LOGGER.info("Processing Request %s" % job.name)
        run = RunFactory.prepare(job)
        t = threading.Thread(target=self._run, args=(run,))
        t.start()
        return run

    def _run(self, run: Run):
        workspace_path = JobService._get_workspace_path()
        JobService._clone_repo(run.job.git_url, workspace_path)
        LOGGER.debug("Cloned Repo into Workspace %s" % workspace_path)
        RunFactory.configure_from_yml_file(run, "%s/%s" % (workspace_path, constants.PIPELINE_YML_NAME))
        LOGGER.info("Starting a new run thread for uuid %s" % run.uuid)
        self.run_service.start(run, workspace_path)
        shutil.rmtree(workspace_path)
        LOGGER.info("Finishing run thread for uuid %s" % run.uuid)

    @staticmethod
    def _get_workspace_path() -> str:
        return "%s/%s/%s" % (
            constants.PIPELINE_BASE,
            constants.PROJECTS_FOLDER,
            uuid.uuid4()
        )

    @staticmethod
    def _clone_repo(url: str, workspace_path: str) -> Repo:
        if os.path.exists(workspace_path): shutil.rmtree(workspace_path)
        return Repo.clone_from(url, workspace_path)



import logging

LOGGER = logging.getLogger(__name__)


class Job:
    def __init__(self, name: str, git_url: str):
        self.name = name
        self.git_url = git_url
        self._links = None

    def to_dict(self):
        return {
            'name': self.name,
            'gitUrl': self.git_url
        }

    @staticmethod
    def of_dict(_dict):
        job = Job(
            _dict["name"],
            _dict["gitUrl"]
        )
        try:
            job._links = _dict["_links"]
        except KeyError:
            LOGGER.info("No Links detected for Job")
        return job



import logging

from docker import DockerClient, from_env
from docker.models.containers import Container
from fxq.core.stereotype import Service

from fxq.ae.agent import constants
from fxq.ae.agent.model.run import Command

LOGGER = logging.getLogger(__name__)


@Service
class DockerService:

    def __init__(self):
        self._docker_client: DockerClient = from_env()

    def list_containers(self):
        return self._docker_client.containers.list()

    def provision(self, name: str, image: str, workspace_path: str = None) -> Container:
        name = name.replace(" ", "_")
        image = image if ":" in image else image + ":latest"
        LOGGER.info("Pulling Container Image %s" % image)
        self._docker_client.images.pull(image)
        LOGGER.info("Provisioning Container \"%s\" with Image:\"%s\"" % (name, image))
        if workspace_path is None:
            return self._provision_without_volume(name, image)
        else:
            return self._provision_with_volume(name, image, workspace_path)

    def run_command(self, container: Container, instruction: str):
        LOGGER.info("Running Container Command \"%s\"" % instruction)
        response = container.exec_run(
            ["/bin/sh", "-c", instruction],
            privileged=True,
            stream=True,
            demux=True,
            workdir=constants.PIPELINE_MOUNT_TARGET
        )
        return response.output

    def teardown(self, container: Container):
        LOGGER.debug("Starting Teardown")
        container.stop(timeout=1)
        container.remove()

    def _provision_without_volume(self, name: str, image: str) -> Container:
        return self._docker_client.containers.run(
            name=name,
            image=image,
            command="tail -f /dev/stdout",
            detach=True
        )

    def _provision_with_volume(self, name: str, image: str, workspace_path: str) -> Container:
        return self._docker_client.containers.run(
            name=name,
            image=image,
            command="tail -f /dev/stdout",
            detach=True,
            volumes={
                workspace_path: {
                    'bind': constants.PIPELINE_MOUNT_TARGET,
                    'mode': 'rw'
                }
            }
        )



import http
import logging
import os
import socket

import requests
from fxq.core.stereotype import Service

from fxq.ae.agent.constants import ANALYTICS_SERVICE_ID

LOGGER = logging.getLogger(__name__)

host_name = socket.gethostname()
host_ip = socket.gethostbyname(host_name)

try:
    _host = os.environ["spring.cloud.consul.discovery.hostname"]
except KeyError:
    _host = host_name if 'fxq.net' in host_name else host_ip
    LOGGER.warning(
        f'Environment Variable: "spring.cloud.consul.discovery.hostname" is not defined, falling back to {_host}')

try:
    _port = int(os.environ["spring.cloud.consul.discovery.port"])
except KeyError:
    _port = 5000
    LOGGER.warning(
        f'Environment Variable: "spring.cloud.consul.discovery.port" is not defined, falling back to {_port}')

try:
    _consul_host = os.environ["spring.cloud.consul.host"]
except KeyError:
    _consul_host = 'localhost'
    LOGGER.warning(f'Environment Variable: "spring.cloud.consul.host is not defined, falling back to {_consul_host}')

try:
    _consul_port = int(os.environ["spring.cloud.consul.port"])
except KeyError:
    _consul_port = 8500
    LOGGER.warning(f'Environment Variable: "spring.cloud.consul.port" is not defined, falling back to {_consul_port}')


@Service
class ConsulService:
    def __init__(self):
        self.host = _host
        self.port = _port
        self.consul_host = _consul_host
        self.consul_port = 8500
        self._register_host()

    def _register_host(self):
        service_def = {
            "ID": "service:ae-agent-5000",
            "Name": "ae-agent",
            "Tags": ["secure=false"],
            "Address": self.host,
            "Port": self.port,
            "Meta": {
            },
            "EnableTagOverride": False,
            "Check": {
                "DeregisterCriticalServiceAfter": "60s",
                "HTTP": f"http://{self.host}:{self.port}/api/actuator/health",
                "Interval": "10s"
            },
            "Weights": {
                "Passing": 10,
                "Warning": 1
            }
        }

        r = requests.put(f'http://{self.consul_host}:{self.consul_port}/v1/agent/service/register', json=service_def)
        if r.status_code != http.HTTPStatus.OK:
            LOGGER.error(f'Registration Failed CODE:{r.status_code} ERROR:{r.text}')
            raise Exception(f'Failed to register with Consul at http://{self.consul_host}:{self.consul_port}')

    def get_callback_host(self):
        r = requests.get(f'http://{self.consul_host}:{self.consul_port}/v1/agent/service/{ANALYTICS_SERVICE_ID}')
        if r.status_code != http.HTTPStatus.OK:
            LOGGER.error(f'Query Failed CODE:{r.status_code} ERROR:{r.text}')
            raise Exception(
                f'Failed to get the Analytics Service using ID "analytics-8200", check it is registered with consul'
            )

        service_definition = r.json()

        return f'http://{service_definition["Address"]}:{service_definition["Port"]}'



import pathlib

from setuptools import setup

setup(
    name='fxq-ae-agent',
    version='0.0.16',
    packages=[
        'fxq.ae.agent',
        'fxq.ae.agent.callback',
        'fxq.ae.agent.factory',
        'fxq.ae.agent.model',
        'fxq.ae.agent.service',
    ],
    url='https://bitbucket.org/fxquants/ae-agent/',
    license='MIT',
    author='Jonathan Turnock',
    author_email='jonathan.turnock@fxquants.net',
    description='Analytics Engine Agent Client, Provisions and executes docker pipelines from git repo yml',
    long_description=(pathlib.Path(__file__).parent / "README.md").read_text(),
    long_description_content_type="text/markdown",
    install_requires=[
        'docker',
        'flask',
        'fxq-core',
        'gitpython',
        'pyyaml'
    ],
    entry_points={
        'console_scripts': ['fxq-ae-agent=fxq.ae.agent.cli:main'],
    }
)



import yaml

from fxq.ae.agent.model.job import Job
from fxq.ae.agent.model.run import Run, Command, Step


class RunFactory:

    @staticmethod
    def prepare(job: Job):
        return Run(job)

    @staticmethod
    def configure_from_yml_file(run: Run, yml_path: str) -> Run:
        with open(yml_path) as ymlf:
            run_yml = yaml.load(ymlf, Loader=yaml.SafeLoader)
            s_no = 0
            for s in run_yml["pipelines"]["steps"]:
                s_no += 1
                step = Step(run, s_no, s["step"]["name"], s["step"]["image"])
                se_no = 0
                for se in s["step"]["script"]:
                    se_no += 1
                    command = Command(step, se_no, se)
                    step.commands.append(command)

                run.steps.append(step)

            return run
